from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional

from pydantic.networks import stricturl

@dataclass
class GCSSourceConfig:
    """
    Attributes:
        template_file_name: a templated file name that applies to all tables. The DAGs will search for one file per table accordingly,
            and no metadata file is required if this field is supplied.
        file_prefix_filtering: option to search for file by prefix (e.g. if the end of file name is a randomly generated string.)
    """
    template_file_name: Optional[str]
    directory_prefix: Optional[str]
    file_prefix_filtering: bool
    date_format: str
    delimeter: str
    source_format: str


from dataclasses import fields
from urllib.parse import urlparse
import logging

from airflow.models.dag import DAG
from airflow.operators.python_operator import PythonOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
from airflow.sensors.external_task import ExternalTaskSensor
from airflow.providers.google.cloud.hooks.kms import CloudKMSHook
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from airflow.models import Variable

from gcp_airflow_foundations.source_class.jdbc_dataflow_source import JdbcToBQDataflowDagBuilder
from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder
import gcp_airflow_foundations.common.dataflow.jdbc.mysql.mysql_query_helpers as mysql_helpers

class MySQLToBQDataflowDagBuilder(JdbcToBQDataflowDagBuilder):
    """
    Builds DAGs to load a CSV file from GCS to a BigQuery Table.
    """
    source_type = "MYSQL"

    def create_job_params(self, config_params, destination_table, table_name, destination_schema_table, query_schema, owner, **kwargs):
        #   1.  Generate SQL Query
        #    a. get source schema from BQ table
        bq_hook = BigQueryHook()
        if not query_schema:
            schema_query = mysql_helpers.get_table_schema_query(destination_schema_table, table_name)
            schema_df = bq_hook.get_pandas_df(schema_query, dialect="standard")

            column_names = schema_df.iloc[:,0]
            data_types = schema_df.iloc[:,1]

            #    b. apply castings if provided   
            casted_columns = mysql_helpers.cast_columns(column_names, data_types, config_params["sql_casts"])
            logging.info(casted_columns)

            #    c. get query
            query = mysql_helpers.get_query_for_mysql_load_full(table_name, casted_columns, owner)
            logging.info(query)
        
        else:
            query = mysql_helpers.get_schema_query(owner)

        # 2. KMS-encrypt credentials for Dataflow job
        logging.info("Encrypting DB credentials with Google Cloud KMS.")

        kms_hook = CloudKMSHook()
        kms_key = config_params["kms_key_path"]
        
        jdbc_url = bytes(config_params["jdbc_url"], 'utf-8')
        jdbc_user = bytes(config_params["jdbc_user"], 'utf-8')
        jdbc_pass = bytes(Variable.get(config_params["jdbc_pass_secret_name"]), 'utf-8')

        enc_url = kms_hook.encrypt(key_name=kms_key, plaintext=jdbc_url)
        enc_user = kms_hook.encrypt(key_name=kms_key, plaintext=jdbc_user)
        enc_pass = kms_hook.encrypt(key_name=kms_key, plaintext=jdbc_pass)

        # Set up parameters to pass to Dataflow job, and XCom push them
        dataflow_default_options = {
            "project": config_params["project"],
           # "region": config_params["region"], 
            "subnetwork": config_params["subnetwork"]
        }
        logging.info(dataflow_default_options)

        ds = kwargs["ds"]

        parameters = {
            "driverClassName": config_params["jdbc_driver_class"],
            "driverJars": config_params["jdbc_jar_path"],
            "KMSEncryptionKey": kms_key,
            "outputTable": destination_table + f"_{ds}",
            "bigQueryLoadingTemporaryDirectory": config_params["bq_load_temp_directory"],
            "connectionURL": enc_url,
            "username": enc_user,
            "password": enc_pass,
            "query": query
        }
        logging.info(parameters)

        kwargs['ti'].xcom_push(key='dataflow_default_options', value=dataflow_default_options)
        kwargs['ti'].xcom_push(key='parameters', value=parameters)

    def get_landing_schema(self, schema_table, source_table):

        bq_hook = BigQueryHook()

        schema_query = mysql_helpers.get_table_schema_query(schema_table, source_table)
        schema_df = bq_hook.get_pandas_df(schema_query, dialect="standard")
        schema_df = schema_df.values.tolist()

        column_names = [x[0] for x in schema_df]
        dtypes = [x[1] for x in schema_df]

        dtypes = mysql_helpers.mysql_to_bq(dtypes)

        schema_fields = []
        for i in range(len(dtypes)):
            schema_fields.append({"name": column_names[i],
                                "type": dtypes[i],
                                "mode": "NULLABLE"})

        logging.info(schema_fields)

        return schema_fields



from dataclasses import fields
from urllib.parse import urlparse
from abc import ABC, abstractmethod, abstractproperty
import logging

from airflow.models.dag import DAG
from airflow.utils.task_group import TaskGroup
from airflow.operators.python_operator import PythonOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator

from gcp_airflow_foundations.base_class.salesforce_ingestion_config import SalesforceIngestionConfig

from gcp_airflow_foundations.operators.api.operators.sf_to_gcs_query_operator import SalesforceToGcsQueryOperator
from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class FTPtoBQDagBuilder(DagBuilder):
    """
    Builds DAGs to load files from FTP to BigQuery.
    """
    source_type = "FTP"

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type     

    def get_bq_ingestion_task(self, dag, table_config):
        taskgroup = TaskGroup(group_id="ftp_taskgroup")

        metadata_sensor_task = self.metadata_file_sensor(table_config, taskgroup)
        schema_sensor_task = self.schema_file_sensor(table_config, taskgroup)
        get_file_list_task = self.get_file_list_task(table_config, taskgroup)
        sensor_task = self.file_sensor(table_config, taskgroup)
        file_ingestion_task = self.file_ingestion_task(table_config, taskgroup)
        load_to_landing_task = self.load_to_landing_task(table_config, taskgroup)

        metadata_sensor_task >> schema_sensor_task >> get_file_list_task >> sensor_task >> file_ingestion_task >> load_to_landing_task

        return taskgroup

    @abstractmethod
    def metadata_file_sensor(self, table_config):
        """
        Implements an Airflow sensor to wait for initial metadata files for ingestion,
        e.g. a metadata file containing a list of files to ingest for .csv files from GCS->BQ,
             .SUCCESS file for a directory of .parquet files from an SFTP server, etc.
        """
        pass

    @abstractmethod
    def schema_file_sensor(self, table_config):
        """
        Implements an Airflow sensor to wait for an (optional) schema file for the table
        """
        pass

    @abstractmethod
    def get_file_list_task(self, table_config):
        """
        Implements a Python operator that should get a list of the files to ingest and xcom push them as follows:
            kwargs['ti'].xcom_push(key='file_list', value=file_list)
        """
        pass

    @abstractmethod
    def file_ingestion_task(self, table_config):
        """
        Implements an Airflow task to ingest the files from the FTP source into GCS (e.g. from an SFTP server or an AWS bucket)
        """
        pass

    @abstractmethod
    def file_sensor(self, table_config):
        """
        Returns an Airflow sensor that waits for the list of files specified the metadata file provided
        Should be Xcom pulled from get_file_list_task()
        """
        pass

    @abstractmethod
    def load_to_landing_task(self, table_config):
        pass

    def validate_extra_options(self):
        tables = self.config.tables


import json
import logging

from airflow.contrib.hooks.bigquery_hook import BigQueryHook


from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults

from gcp_airflow_foundations.operators.api.hooks.twilio_hook import TwilioHook


class TwilioToBigQueryOperator(BaseOperator):
    """
    Twilio Text Message data to BigQuery
    
    Attributes:
        twilio_account_sid: Twilio Account SID to be used in the API
        project_id: GCP project ID               
        table_id: Staging table name    
        dataset_id: Staging dataset name
        schema_fields: Schema fields provided by the Airflow Framework for the Twilio Staging table
        delegate_to: The account to impersonate using domain-wide delegation of authority, if any
        gcp_conn_id: Airflow GCP connection ID
        twilio_conn_id: Airflow Twilio connection ID
        column_mapping: Column mapping dictionary
    """
    template_fields = ('dataset_id', 'table_id', 'project_id', 'labels')

    # pylint: disable=too-many-arguments
    @apply_defaults
    def __init__(self,
                 twilio_account_sid,
                 dataset_id,
                 table_id,
                 project_id=None,
                 schema_fields=None,
                 gcs_schema_object=None,
                 time_partitioning=None,
                 gcp_conn_id='google_cloud_default',
                 twilio_conn_id='twilio_default',
                 delegate_to=None,
                 labels=None,
                 encryption_configuration=None,
                 *args, **kwargs):

        super(TwilioToBigQueryOperator, self).__init__(*args, **kwargs)

        self.twilio_account_sid = twilio_account_sid
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.schema_fields = schema_fields
        self.gcs_schema_object = gcs_schema_object
        self.gcp_conn_id = gcp_conn_id
        self.twilio_conn_id = twilio_conn_id
        self.delegate_to = delegate_to
        self.time_partitioning = {} \
            if time_partitioning is None else time_partitioning
        self.labels = labels
        self.encryption_configuration = encryption_configuration


    def execute(self, context):
        twilio_hook = TwilioHook(twilio_conn_id=self.twilio_conn_id)
        
        bq_hook = BigQueryHook(gcp_conn_id=self.gcp_conn_id)
        conn = bq_hook.get_conn()
        cursor = conn.cursor()

        if not bq_hook.table_exists(project_id=self.project_id,
                                dataset_id=self.dataset_id,
                                table_id=self.table_id
                                ):

            if not self.schema_fields and self.gcs_schema_object:

                parsed_url = urlparse(self.gcs_schema_object)
                gcs_bucket = parsed_url.netloc
                gcs_object = parsed_url.path.lstrip('/')

                gcs_hook = GoogleCloudStorageHook(
                    google_cloud_storage_conn_id=self.google_cloud_storage_conn_id,
                    delegate_to=self.delegate_to)
                schema_fields = json.loads(gcs_hook.download(
                    gcs_bucket,
                    gcs_object).decode("utf-8"))
            else:
                schema_fields = self.schema_fields
            
            cursor.create_empty_table(
                project_id=self.project_id,
                dataset_id=self.dataset_id,
                table_id=self.table_id,
                schema_fields=schema_fields,
                time_partitioning=self.time_partitioning,
                labels=self.labels,
                encryption_configuration=self.encryption_configuration
            )

        service = bq_hook.get_service()

        body = {
            'rows' : twilio_hook.get_text_messages(account_sid=self.twilio_account_sid)
        }

        service_exec = service.tabledata().insertAll(
            body = body,
            projectId = self.project_id,
            tableId = self.table_id,
            datasetId = self.dataset_id
        )

        results = service_exec.execute()

        logging.info(results)


from airflow import DAG
from airflow.exceptions import AirflowException

from airflow.utils.task_group import TaskGroup

from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryCreateEmptyTableOperator,
    BigQueryExecuteQueryOperator
)

from gcp_airflow_foundations.operators.gcp.ods.ods_merge_table_operator import MergeBigQueryODS
from gcp_airflow_foundations.operators.gcp.schema_migration.schema_migration_operator import MigrateSchema

from gcp_airflow_foundations.operators.gcp.create_table import CustomBigQueryCreateEmptyTableOperator


def ods_builder(
    project_id,
    table_id,
    dataset_id,
    landing_zone_dataset,
    landing_zone_table_name_override,
    column_mapping,
    surrogate_keys,
    ingestion_type,
    ods_table_config,
    location,
    dag,
    time_partitioning=None,
    labels=None,
    encryption_configuration=None) -> TaskGroup:

    """
    Method for returning a Task Group for 1) creating an empty target ODS table (if it doesn't already exist) and for 2) for merging the staging table data into the target ODS table
    """
    taskgroup = TaskGroup(group_id="create_ods_merge_taskgroup")

    #1 Check if ODS table exists and if not create an empty table
    create_table = CustomBigQueryCreateEmptyTableOperator(
        task_id="create_ods_table",
        project_id=project_id,
        dataset_id=dataset_id,
        table_id=table_id,
        time_partitioning=None,
        task_group=taskgroup,
        dag=dag
    )
    
    #2 Migrate schema
    migrate_schema = MigrateSchema(
        task_id="schema_migration",
        project_id=project_id,
        table_id=table_id,
        dataset_id=dataset_id, 
        task_group=taskgroup,
        dag=dag
    )
    
    #3 Merge or truncate tables based on the ingestion type defined in the config file and insert metadata columns
    insert = MergeBigQueryODS(
        task_id="insert_into_ods_table",
        project_id=project_id,
        stg_dataset_name=landing_zone_dataset,
        data_dataset_name=dataset_id,
        stg_table_name=landing_zone_table_name_override,
        data_table_name=table_id,
        surrogate_keys=surrogate_keys,
        column_mapping=column_mapping,
        ingestion_type=ingestion_type,
        ods_table_config=ods_table_config,
        location=location,
        task_group=taskgroup,
        dag=dag
    )

    create_table >> migrate_schema >> insert

    return taskgroup



import paramiko
import socket
import os
from stat import S_ISDIR

class SFTPSession(object):
    def __init__(self,hostname,username='root',key_file=None,password=None):
        transport = paramiko.Transport((hostname, 22))
        mykey = paramiko.RSAKey.from_private_key_file(key_file)
        transport.connect(username= username, pkey=mykey)
        self.sftp = paramiko.SFTPClient.from_transport(transport)
        
    def command(self,cmd):
        #  Breaks the command by lines, sends and receives 
        #  each line and its output separately
        #
        #  Returns the server response text as a string
        
        chan = self.t.open_session()
        chan.get_pty()
        chan.invoke_shell()
        chan.settimeout(20.0)
        ret=''
        try:
            ret+=chan.recv(1024)
        except:
            chan.send('\n')
            ret+=chan.recv(1024)
        for line in cmd.split('\n'):
            chan.send(line.strip() + '\n')
            ret+=chan.recv(1024)
        return ret

    def get_stats(self, remotefile):
        return self.sftp.stat(remotefile)

    def put(self,localfile,remotefile):
        #  Copy localfile to remotefile, overwriting or creating as needed.
        self.sftp.put(localfile,remotefile)
    
    def remotepath_join(self,*args):
        #  Bug fix for Windows clients, we always use / for remote paths
        return '/'.join(args)
    
    
    def put_all(self,localpath,remotepath):
        #  recursively upload a full directory
        os.chdir(os.path.split(localpath)[0])
        parent=os.path.split(localpath)[1]
        for path,_,files in os.walk(parent):
            try:
                self.sftp.mkdir(self.remotepath_join(remotepath,path))
            except:
                pass
            for filename in files:
                self.put(os.path.join(path,filename),self.remotepath_join(remotepath,path,filename))
    
    def get(self,remotefile,localfile):
        #  Copy remotefile to localfile, overwriting or creating as needed.
        self.sftp.get(remotefile,localfile)
    
    def sftp_walk(self,remotepath):
        # Kindof a stripped down  version of os.walk, implemented for 
        # sftp.  Tried running it flat without the yields, but it really
        # chokes on big directories.
        path=remotepath
        files=[]
        folders=[]
        for f in self.sftp.listdir_attr(remotepath):
            if S_ISDIR(f.st_mode):
                folders.append(f.filename)
            else:
                files.append(f.filename)
        yield path,folders,files
        for folder in folders:
            new_path=self.remotepath_join(remotepath,folder)
            for x in self.sftp_walk(new_path):
                yield x
        
    def get_all(self,remotepath,localpath):
        #  recursively download a full directory
        #  Harder than it sounded at first, since paramiko won't walk
        #
        # For the record, something like this would generally be faster:
        # ssh user@host 'tar -cz /source/folder' | tar -xz
        
        self.sftp.chdir(os.path.split(remotepath)[0])
        parent=os.path.split(remotepath)[1]
        try:
            os.mkdir(localpath)
        except FileExistsError:
            pass
        for path,_,files in self.sftp_walk(parent):
            try:
                os.mkdir(self.remotepath_join(localpath,path))
            except FileExistsError:
                pass
            for filename in files:
                print(self.remotepath_join(path,filename),os.path.join(localpath,path,filename))
                self.get(self.remotepath_join(path,filename),os.path.join(localpath,path,filename))
    def write_command(self,text,remotefile):
        #  Writes text to remotefile, and makes remotefile executable.
        #  This is perhaps a bit niche, but I was thinking I needed it.
        #  For the record, I was incorrect.
        self.sftp.open(remotefile,'w').write(text)
        self.sftp.chmod(remotefile,755)


from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional

@dataclass
class OdsTableMetadataConfig:
    """
    Attributes:
        hash_column_name : Hash column name
        primary_key_hash_column_name: Primary key hash column name
        ingestion_time_column_name : Insertion time column name
        update_time_column_name : Update time column name
    """
    hash_column_name: str = 'af_metadata_row_hash'
    primary_key_hash_column_name: str = 'af_metadata_primary_key_hash'
    ingestion_time_column_name: str = 'af_metadata_inserted_at'
    update_time_column_name: str = 'af_metadata_updated_at'



from abc import ABC, abstractmethod, abstractproperty
import logging
from re import A
import re
from airflow.operators.dummy import DummyOperator
from dacite import from_dict
from dataclasses import dataclass

from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.base_class.source_table_config import SourceTableConfig
from gcp_airflow_foundations.base_class.dataflow_job_config import DataflowJobConfig

from dataclasses import fields
from urllib.parse import urlparse
import logging

from airflow.models.dag import DAG
from airflow.operators.python_operator import PythonOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
from airflow.sensors.external_task import ExternalTaskSensor
from airflow.providers.google.cloud.operators.dataflow import DataflowTemplatedJobStartOperator
from airflow.providers.google.cloud.operators.bigquery import BigQueryCreateEmptyTableOperator
from airflow.providers.google.cloud.hooks.kms import CloudKMSHook
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from airflow.models import Variable
from airflow.operators.subdag import SubDagOperator

from gcp_airflow_foundations.common.dataflow.jdbc.dataflow_taskgroups import dataflow_taskgroup_builder

from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class JdbcToBQDataflowDagBuilder(DagBuilder):

    source_type = "JDBC"

    def get_extra_dags(self):
        schema_dag = self.get_schema_dag()
        if schema_dag:
            return [schema_dag]
        else:
            return schema_dag

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type     

    def get_bq_ingestion_task(self, dag, table_config):
        data_source = self.config.source

        # Source level Parameters
        system_name = data_source.extra_options["dataflow_job_config"]["system_name"]
        gcp_project = data_source.gcp_project
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset

        # Table level parameters
        dataflow_job_params = data_source.extra_options["dataflow_job_config"]
        schema_table = dataflow_job_params["bq_schema_table"]
        ingest_metadata = dataflow_job_params["ingest_metadata"]
        table_name = table_config.landing_zone_table_name_override
        destination_table = f"{gcp_project}:{landing_dataset}.{table_name}"
        destination_schema_table = f"{gcp_project}.{landing_dataset}.{schema_table}"

        if "table_type_casts" in data_source.extra_options["dataflow_job_config"]:
            table_type_casts = data_source.extra_options["dataflow_job_config"]["table_type_casts"]
        else:
            table_type_casts = {}

        taskgroup = dataflow_taskgroup_builder(
            query_schema=False,
            dataflow_job_params=dataflow_job_params,
            destination_table=destination_table,
            destination_schema_table=destination_schema_table,
            table_name=table_name,
            system_name=system_name,
            create_job_params=self.create_job_params,
            run_dataflow_job=self.run_dataflow_job,
            create_table=self.create_table,
            ingest_metadata=ingest_metadata,
            table_type_casts=table_type_casts
        )

        return taskgroup
    
    def get_schema_dag(self):
        """
        This method returns a singular dag that runs a Dataflow job to fetch the global schemas from source.
        """
        data_source = self.config.source
        gcp_project = data_source.gcp_project
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset

        dataflow_job_params = data_source.extra_options["dataflow_job_config"]
        ingest_metadata = dataflow_job_params["ingest_metadata"]
        schema_table_name = dataflow_job_params["bq_schema_table"]
        destination_table = f"{gcp_project}:{landing_dataset}.{schema_table_name}"
        system_name = dataflow_job_params["system_name"]

        if ingest_metadata:
            with DAG(
                dag_id=f"{system_name}_upload_schema",
                description=f"Upload source schemas for all tables to BQ",
                schedule_interval="@daily",
                default_args=self.default_task_args_for_table(
                    self.config, self.config.tables[0]
                )
            ) as schema_dag:

                taskgroup = dataflow_taskgroup_builder(
                #  schema_dag,
                    query_schema=True,
                    dataflow_job_params=dataflow_job_params,
                    destination_table=destination_table,
                    destination_schema_table=f"{gcp_project}.{landing_dataset}.{schema_table_name}",
                    table_name=schema_table_name,
                    system_name=system_name,
                    create_job_params=self.create_job_params,
                    run_dataflow_job=self.run_dataflow_job,
                    create_table=self.create_table,
                    ingest_metadata=ingest_metadata,
                    table_type_casts={}
                )
                taskgroup.dag = schema_dag

                return schema_dag
        else:
            return

    def run_dataflow_job(self, template_path, system_name, table_name, query_schema, **kwargs):
        ti = kwargs['ti']
        xcom_task_pickup = "dataflow_taskgroup.create_job_parameters"

        dataflow_default_options = ti.xcom_pull(key='dataflow_default_options', task_ids=xcom_task_pickup)
        parameters = ti.xcom_pull(key='parameters', task_ids=xcom_task_pickup)
        job_name = f"{system_name.lower()}-upload-{table_name.lower()}-to-bq".replace("_", "-")

        # hacky - TO FIX
        if isinstance(parameters["query"], list):
            parameters["query"] = parameters["query"][0]

        trigger_job = DataflowTemplatedJobStartOperator(
            task_id=job_name,
            job_name=job_name,
            template=template_path,
            dataflow_default_options=dataflow_default_options,
            parameters=parameters
        )
        trigger_job.execute(context=kwargs)
    
    def create_table(self, destination_table, schema_table, source_table, table_type_casts, **kwargs):
        ds = kwargs["ds"]
        ids = destination_table.split(":")
        project_id = ids[0]
        dataset_id = ids[1].split(".")[0]
        table_id = ids[1].split(".")[1] + f"_{ds}"

        logging.info(ids)
        bq_hook = BigQueryHook()
        table_exists = bq_hook.table_exists(dataset_id=dataset_id, table_id=table_id)

        schema_fields = self.get_landing_schema(schema_table, source_table)

        logging.info(schema_fields)
        for key, val in table_type_casts.items():
            for i in range(len(schema_fields)):
                if schema_fields[i]["type"] is key:
                    logging.info(key, val)
                    schema_fields[i]["type"] = val
        logging.info(schema_fields)

        if not table_exists:
            create_table_op = BigQueryCreateEmptyTableOperator(
                task_id="create_table",
                project_id=project_id,
                dataset_id=dataset_id,
                table_id=table_id,
                schema_fields=schema_fields
            )
            create_table_op.execute(context=kwargs)
        
    @abstractmethod
    def create_job_params(self, config_params, destination_table, **kwargs):
        """
        Inputs:
        config_params:
            dictionary corresponding to a DataflowJobConfig
        destination_table:
            Bigquery table in form {gcp_project}.{bq_dataset}.{table_name}
        query_schema:
            boolean - whether to query the schema or not

        When implemented, this method should create the following dictionaries, fill them 
        with the required parameters, and XCom push them

        DATAFLOW_DEFAULT_OPTIONS:
           project
           region
           subnetwork
        PARAMETERS:
           driverClassName
           driverJars
           KMSEncryptionKey 
           outputTable
           bigQueryLoadingTemporaryDirectory
           connectionURL 
           username
           password 
           query
        """
        pass

    @abstractmethod
    def get_landing_schema(self, schema_table, source_table):
        return None

    def validate_extra_options(self):
        # try and parse as DataflowJobConfig
        job_cfg = from_dict(data_class=DataflowJobConfig, data=self.config.source.extra_options["dataflow_job_config"])
        
        # For tables - assert only FULL ingestion available for now   
        tables = self.config.tables
       # for table in tables:
       #     assert table.ingestion_type == "FULL"

            



import json
import logging

from airflow.contrib.hooks.bigquery_hook import BigQueryHook
from airflow.contrib.hooks.gcs_hook import (
    GoogleCloudStorageHook
)

from urllib.parse import urlparse


def read_schema_from_gcs(
    gcs_schema_object,
    google_cloud_storage_conn_id='google_cloud_default',
    bigquery_conn_id='google_cloud_default',
    **kwargs) -> list:
    """
        Helper method to load table schema from a GCS URI
    """
    
    bq_hook = BigQueryHook(bigquery_conn_id=bigquery_conn_id, delegate_to=None)

    parsed_url = urlparse(gcs_schema_object)
    gcs_bucket = parsed_url.netloc
    gcs_object = parsed_url.path.lstrip('/')

    gcs_hook = GoogleCloudStorageHook(
        google_cloud_storage_conn_id=google_cloud_storage_conn_id,
        delegate_to=None)

    schema_fields = json.loads(gcs_hook.download(
        gcs_bucket,
        gcs_object).decode("utf-8"))

    return schema_fields


import requests
import time
from datetime import datetime
import json
from requests.auth import HTTPBasicAuth

import hashlib

from airflow.hooks.base_hook import BaseHook

class TwilioHook(BaseHook):
    """
    Airflow Hook to connect to Twilio's API and retrieve Text Message data
    """
    def __init__(self, twilio_conn_id='twilio_default') -> None:
        self.twilio_conn_id = twilio_conn_id,
        self.base_uri = 'https://api.twilio.com/2010-04-01'

        connection = self.get_connection(twilio_conn_id)
        self.sid = connection.login
        self.secret = connection.password

    def transform_text_messages(
        self,
        text_messages
    ) -> list:

        rows = []
        for i in text_messages:
            i['feedback'] = i['subresource_uris']['feedback'] 
            i['media'] = i['subresource_uris']['media']
            i.pop('subresource_uris', None)

            i['date_created'] = datetime.strptime(i['date_created'], '%a, %d %b %Y %H:%M:%S %z').strftime('%Y-%m-%d %H:%M:%S')
            i['date_sent'] = datetime.strptime(i['date_sent'], '%a, %d %b %Y %H:%M:%S %z').strftime('%Y-%m-%d %H:%M:%S')
            i['date_updated'] = datetime.strptime(i['date_updated'], '%a, %d %b %Y %H:%M:%S %z').strftime('%Y-%m-%d %H:%M:%S')

            keys = ['sid']

            joined_keys = ''.join([i[key] for key in keys])

            rows.append(
                {
                    'insertId':hashlib.md5(joined_keys.encode('utf-8')).hexdigest(),
                    'json':i
                    }
            )
        
        return rows

    def get_text_messages(
        self,
        account_sid,
        FromDate=None,
        ToDate=None
    ) -> list:

        URL = f'{self.base_uri}/Accounts/{account_sid}/Messages.json'

        auth = HTTPBasicAuth(self.sid, self.secret)

        params = {
            'PageSize':1000
        }

        if FromDate:
            params['DateSent>'] = FromDate

        if ToDate:
            params['DateSent<'] = ToDate

        text_messages = []

        while True:
            time.sleep(1)

            r = requests.get(URL, params=params, auth=auth)
            r.raise_for_status()

            body = r.json()
            text_messages.extend(body['messages'])

            NextPageURI = body.get('next_page_uri', None)
            if not NextPageURI:
                break

            URL = f'https://api.twilio.com/{NextPageURI}'

        rows = self.transform_text_messages(text_messages)

        return rows


from dataclasses import fields
from os import X_OK
from urllib.parse import urlparse
from abc import ABC, abstractmethod, abstractproperty
import logging
import json
import pandas as pd
from dacite import from_dict
from dataclasses import dataclass
from datetime import datetime

from airflow.models.dag import DAG
from airflow.utils.task_group import TaskGroup
from airflow.operators.python_operator import PythonOperator
from airflow.operators.dummy import DummyOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
from airflow.providers.google.cloud.sensors.gcs import GCSObjectExistenceSensor
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.exceptions import AirflowException

from gcp_airflow_foundations.base_class.gcs_table_config import GCSTableConfig
from gcp_airflow_foundations.base_class.gcs_source_config import GCSSourceConfig
from gcp_airflow_foundations.operators.api.operators.sf_to_gcs_query_operator import SalesforceToGcsQueryOperator
from gcp_airflow_foundations.operators.api.sensors.gcs_sensor import GCSObjectListExistenceSensor
from gcp_airflow_foundations.operators.api.sensors.gcs_prefix_sensor import GCSObjectPrefixListExistenceSensor
from gcp_airflow_foundations.source_class.ftp_source import FTPtoBQDagBuilder
from gcp_airflow_foundations.base_class.data_source_table_config import DataSourceTablesConfig
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder


class GCSFiletoBQDagBuilder(FTPtoBQDagBuilder):
    """
    Builds DAGs to load files from GCS to a BigQuery Table.

    For GCS->BQ ingestion, either a metadata file is required or the field templated_file_name must be provided. 
    If a metadata file is provided, itt can be a fixed file, or can be a new file supplied daily.
    Airflow context variables are supported for the file naming, e.g.
        TABLE_METADATA_FILE_{{ ds }}.csv
    for a metadata file supplied daily.

    The format of the metadata file should be a csv with one column as follows:
        FILE_NAME_1
        ...
        FILE_NAME_N
    with all files to ingest
    """
    source_type = "GCS"
    gcs_hook = GCSHook()

    def metadata_file_sensor(self, table_config, taskgroup):
        """
        Implements a sensor for the metadata file specified in the table config.
        """
        if "metadata_file" in table_config.extra_options.get("gcs_table_config"):
            metadata_file_name = table_config.extra_options.get("gcs_table_config")["metadata_file"]
            bucket = self.config.source.extra_options["gcs_bucket"]
            return GCSObjectExistenceSensor(
                task_id="wait_for_metadata_file",
                bucket=bucket,
                object=metadata_file_name,
                task_group=taskgroup
            )
        elif "sensor_file" in table_config.extra_options.get("gcs_table_config"):
            metadata_file_name = table_config.extra_options.get("gcs_table_config")["metadata_file"]
            bucket = self.config.source.extra_options["gcs_bucket"]
            return self.get_sftp_sensor(table_config, taskgroup)
        else:
            return DummyOperator(
                task_id="dummy_metadata_file_ingestion",
                task_group=taskgroup
            )

    def file_ingestion_task(self, table_config, taskgroup):
        """
        Ingest from SFTP
        """
        return self.get_sftp_ingestion_operator(table_config, taskgroup)

    def schema_file_sensor(self, table_config, taskgroup):
        """
        Implements an Airflow sensor to wait for an (optional) schema file in GCS
        """
        bucket = self.config.source.extra_options["gcs_bucket"]
        schema_file_name = None
        if "schema_file" in table_config.extra_options.get("gcs_table_config"):
            schema_file_name = table_config.extra_options.get("gcs_table_config")["schema_file"]
        logging.info(schema_file_name)

        if schema_file_name:
            return GCSObjectExistenceSensor(
                task_id="wait_for_schema_file",
                bucket=bucket,
                object=schema_file_name,
                task_group=taskgroup
            )
        else:
            return DummyOperator(
                task_id="dummy_schema_file_ingestion_operator",
                task_group=taskgroup
            )

    def get_file_list_task(self, table_config, taskgroup):
        if "metadata_file" in table_config.extra_options.get("gcs_table_config"):
            execution_type = "metadata"
        else:
            execution_type = "success"

        return PythonOperator(
            task_id="get_file_list",
            op_kwargs={"table_config": table_config,
                        "execution_type": execution_type},
            python_callable=self.get_list_of_files,
            task_group=taskgroup
        )

    def file_sensor(self, table_config, taskgroup):
        """
        Returns an Airflow sensor that waits for the list of files specified by the metadata file provided.
        """
        bucket = self.config.source.extra_options["gcs_bucket"]
        files_to_wait_for = "{{ ti.xcom_pull(key='file_list', task_ids='ftp_taskgroup.get_file_list') }}"

        if self.config.source.extra_options["gcs_source_config"]["file_prefix_filtering"]:
            return GCSObjectPrefixListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                prefixes=files_to_wait_for,
                task_group=taskgroup
            )
        else:
            return GCSObjectListExistenceSensor(
                task_id="wait_for_files_to_ingest",
                bucket=bucket,
                objects=files_to_wait_for,
                task_group=taskgroup
            )

    def load_to_landing_task(self, table_config, taskgroup):
        return PythonOperator(
            task_id="load_gcs_to_landing_zone",
            op_kwargs={"table_config": table_config},
            python_callable=self.load_to_landing_py_op_task,
            task_group=taskgroup
        )   

    def load_to_landing_py_op_task(self, table_config, **kwargs):
        data_source = self.config.source
        bucket = self.config.source.extra_options["gcs_bucket"]

        ti = kwargs['ti']
        ds = kwargs['ds']

        files_to_load = ti.xcom_pull(key='file_list', task_ids='ftp_taskgroup.get_file_list')
        logging.info(files_to_load)

        if self.config.source.extra_options["gcs_source_config"]["file_prefix_filtering"]:
            logging.info("HELLO")
            for i in range(len(files_to_load)):
                matching_gcs_files = self.gcs_hook.list(bucket_name=bucket, prefix=files_to_load[i]) 
                if len(matching_gcs_files) > 1:
                    raise AirflowException(f"There is more than one matching file with the prefix {files_to_load[i]} in the bucket {bucket}")
                files_to_load[i] = matching_gcs_files[0]

        logging.info(files_to_load)

        # Parameters
        bucket = self.config.source.extra_options["gcs_bucket"]
        source_format = self.config.source.extra_options["gcs_source_config"]["source_format"]
        field_delimeter = self.config.source.extra_options["gcs_source_config"]["delimeter"]
        gcp_project = data_source.gcp_project
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset 
        destination_table = f"{gcp_project}:{landing_dataset}.{table_config.landing_zone_table_name_override}" + f"_{ds}"

        schema_file_name = None
        if "schema_file" in table_config.extra_options.get("gcs_table_config"):
            schema_file_name = table_config.extra_options.get("gcs_table_config")["schema_file"]
        logging.info(field_delimeter)
        
        allow_quoted_newlines = False
        if "allow_quoted_newlines" in table_config.extra_options.get("gcs_table_config"):
           allow_quoted_newlines = table_config.extra_options.get("gcs_table_config")["allow_quoted_newlines"]

        # Get files to load from metadata file
        if schema_file_name:
            schema_file = self.gcs_hook.download(bucket_name=bucket, object_name=schema_file_name)

            # Only supports json schema file format - add additional support if required
            schema_fields = json.loads(schema_file)
            logging.info(schema_fields)
            gcs_to_bq = GCSToBigQueryOperator(
                task_id='import_files_to_bq_landing',
                bucket=bucket,
                source_objects=files_to_load,
                source_format=source_format,
                schema_fields=schema_fields,
                field_delimiter=field_delimeter,
                destination_project_dataset_table=destination_table,
                allow_quoted_newlines=allow_quoted_newlines,
                write_disposition='WRITE_TRUNCATE',
                create_disposition='CREATE_IF_NEEDED',
                skip_leading_rows=1,
            )
        else:
            gcs_to_bq = GCSToBigQueryOperator(
                task_id='import_files_to_bq_landing',
                bucket=bucket,
                source_objects=files_to_load,
                source_format=source_format,
                field_delimiter=field_delimeter,
                destination_project_dataset_table=destination_table,
                allow_quoted_newlines=allow_quoted_newlines,
                write_disposition='WRITE_TRUNCATE',
                create_disposition='CREATE_IF_NEEDED',
                skip_leading_rows=1,
            )
        gcs_to_bq.execute(context=kwargs)

    def get_list_of_files(self, table_config, **kwargs):
        # XCom push the list of files
        bucket = self.config.source.extra_options["gcs_bucket"]
        if "metadata_file" in table_config.extra_options.get("gcs_table_config"):
            metadata_file_name = table_config.extra_options.get("gcs_table_config")["metadata_file"]
            metadata_file = self.gcs_hook.download(bucket_name=bucket, object_name=metadata_file_name, filename="metadata.csv")
            file_list = []
            with open('metadata.csv', newline='') as f:
                for line in f:
                    file_list.append(line.strip())
        else:
            templated_file_name = self.config.source.extra_options["gcs_source_config"]["templated_file_name"]
            templated_file_name = templated_file_name.replace("{{ TABLE_NAME }}", table_config.table_name)
            file_list = [templated_file_name]

        # support replacing files with current dates
        ds = kwargs["ds"]
        ds = datetime.strptime(ds, "%Y-%m-%d").strftime(self.config.source.extra_options["gcs_source_config"]["date_format"])
        file_list[:] = [file.replace("{{ ds }}", ds) if "{{ ds }}" in file else file for file in file_list]

        # add dir prefix to files
        x = "directory_prefix" in self.config.source.extra_options["gcs_source_config"]
        logging.info(x)
        if "directory_prefix" in self.config.source.extra_options["gcs_source_config"]:
            dir_prefix = self.config.source.extra_options["gcs_source_config"]["directory_prefix"]
            file_list[:] = [dir_prefix + file for file in file_list]
        logging.info(file_list)

        kwargs['ti'].xcom_push(key='file_list', value=file_list)

    def validate_extra_options(self):
        pass
        #gcs_source_cfg = from_dict(data_class=GCSSourceConfig, data=self.config.source.extra_options["gcs_source_config"])
        #tables = self.config.tables
        #for table in tables:
        #    gcs_table_cfg = from_dict(data_class=GCSTableConfig, data=table.extra_options.get("gcs_table_config"))


import os
import sys

from setuptools import setup, find_packages
from pathlib import Path

here = os.path.abspath(os.path.dirname(__file__))
about = {}

with open(os.path.join(here, "src", "gcp_airflow_foundations", "__version__.py"), "r") as f:
    exec(f.read(), about)

with open(os.path.join(here, "requirements.txt"), "r") as f:
	requirements = f.read().strip().split('\n')

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

def main():
    metadata = dict(
        name=about["__title__"],
        version=about["__version__"],
        description=about["__description__"],
        long_description=long_description,
        long_description_content_type='text/markdown',
        url=about["__url__"],
        download_url=about["__download_url__"],
        author=about["__author__"],
        author_email=about["__author_email__"],
        license=about["__license__"],
        packages=find_packages(where="src", exclude=("tests")),
        package_dir={"":"src"},
        install_requires=requirements,
        classifiers=[
            'Development Status :: 4 - Beta',      
            'Intended Audience :: Developers',      
            'Topic :: Software Development :: Libraries',
            'License :: OSI Approved :: Apache Software License',   
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ]
    )

    setup(**metadata)


if __name__ == "__main__":
    main()



from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional

@dataclass
class GCSTableConfig:
    """
    Attributes:
        metadata_file: the relative path to the metadata file within the GSC bucket
        schema_file: the (optional) relative path to the schema file for the table
        source_format: file format for all files in this metadata file
    """
    metadata_file: Optional[str]
    schema_file: Optional[str]
    source_format: Optional[str]


from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional

@dataclass
class DataflowJobConfig:
    """
    Attributes:
        project: the GCP project in which the Dataflow job runs
        region: the region in which the Dataflow job should run
        subnetwork: the specific subnetwork in which the Dataflow job should run
        bq_load_temp_directory: GCS directory for loading temporary Dataflow files
        template_path: GCS path to Dataflow template
        jdbc_driver_class: the name of the JDBC driver class to use (e.g. oracle.jdbc.driver.OracleDriver)
        jdbc_jar_path: the GCS path to the driver .jar file
        jdbc_url: a valid JDBC url for connecting to the database
        jdbc_user: the database username
        jdbc_pass_secret_name: the secret name of the database password
        kms_key_path: the KMS key path for encrypting/decrypting JDBC credentials
        sql_casts: a dictionary of sql casts to use when querying the source DB
        database_owner: owner of the tables to query (query scope)
        connection_pool: name of the connection pool to use in Airflow for dataflow
        ingest_metadata: whether to query metadata from the source DB: if False, then a table must be manually created in BQ
    """
    system_name: str
    project: str
    region: str
    subnetwork: str
    bq_load_temp_directory: str
    template_path: str
    jdbc_driver_class: str
    jdbc_jar_path: str
    jdbc_url: str
    jdbc_user: str
    jdbc_pass_secret_name: str
    kms_key_path: str
    sql_casts: Optional[dict]
    bq_schema_table: str
    database_owner: str
    connection_pool: str
    ingest_metadata: bool